import fs from 'fs';
import path from 'path';
import { EventEmitter } from 'events';
import { getModuleHash } from '../util/fileUtil.js';
import { startWatch } from '../watcher/index.js';
import { getWorkspaceScanner } from './workspaceScanner/index.js';
export class ModuleGraph extends EventEmitter {
    constructor(config, logger) {
        super();
        /**
         * All LWC modules and their parts with absolute file paths, for the current workspace.
         */
        this.moduleMap = new Map();
        // Mapping from <absolute file path> => "<namespace>/<componentname>/component.js"
        this.fileToModulePathMap = Object.create(null);
        // Mapping from "<namespace>/<componentname>/component.js" => { moduleFilePath, moduleSourceHash }
        this.modulePathToModuleInfo = Object.create(null);
        this.watcher = startWatch(config.paths);
        this.logger = logger;
        this.workspaceScanner = getWorkspaceScanner(config, logger);
        // init() may be called multiple times in the future(when a directory is deleted or addeds)
        this.init();
    }
    init() {
        this.moduleMap = this.workspaceScanner.getModulesMap();
        if (this.moduleMap.size) {
            this.logger.debug(`[Module Graph] Found ${this.moduleMap.size} LWC components in the current workspace.`);
        }
        else {
            this.logger.debug(`[Module Graph] Found no LWC components in the current workspace.`);
        }
        this.moduleMap.forEach(({ moduleParts }) => {
            Object.assign(this.modulePathToModuleInfo, moduleParts);
            for (const [modulePath, moduleInfo] of Object.entries(moduleParts)) {
                this.fileToModulePathMap[moduleInfo.moduleFilePath] = modulePath;
            }
        });
        this.watcher.on('change', this.fileChangeHandler.bind(this));
        this.watcher.on('error', (event) => {
            this.logger.error('[LWC Dev Server] Encountered error while watching files');
            this.logger.debug(`[LWC Dev Server] File watcher error: ${event.message}`);
            this.emit('close');
        });
    }
    // Given absolute file path, get the modulePath
    // Mapping from <absolute file path> => "<namespace>/<componentname>/component.js"
    getModulePathByFile(path) {
        return this.fileToModulePathMap[path];
    }
    // Given the modulePath, get the absolute file path for that module
    getFileByModulePath(modulePath) {
        return this.modulePathToModuleInfo[modulePath]?.moduleFilePath;
    }
    /**
     * Gets the stale modules from the list of modules
     * @param modules
     * @returns Modules paths that are determined to be stale wrt the content monitored by the dev server
     */
    getStaleModules(modules) {
        const staleModules = [];
        const unTrackedModules = [];
        modules.forEach((mod) => {
            const { modulePath, hash } = mod;
            if (!modulePath.trim())
                return;
            const moduleInfo = this.modulePathToModuleInfo[modulePath];
            if (!moduleInfo) {
                unTrackedModules.push(modulePath);
            }
            else if (moduleInfo.moduleSourceHash !== hash) {
                staleModules.push(modulePath);
            }
        });
        unTrackedModules.length &&
            this.logger.debug(`[Module Graph] Following modules are unknown to dev server: \n\t ${unTrackedModules.join(',\n\t')}`);
        staleModules.length &&
            this.logger.debug(`[Module Graph] Following modules have a newer version: \n\t ${staleModules.join(',\n\t')}`);
        const staleModuleBoundaries = new Set();
        for (const path of staleModules) {
            staleModuleBoundaries.add(this.getModuleBoundary(path));
        }
        return staleModuleBoundaries;
    }
    fileChangeHandler(filePath) {
        const changedFile = path.resolve(filePath);
        if (!fs.lstatSync(changedFile).isFile())
            return;
        const modulePath = this.getModulePathByFile(changedFile);
        if (!modulePath)
            return;
        const moduleInfo = this.modulePathToModuleInfo[modulePath];
        const newModuleHash = getModuleHash(changedFile);
        if (moduleInfo.moduleSourceHash === newModuleHash)
            return;
        this.logger.info(`[LWC Dev Server] File '${modulePath}' modified - updating.`);
        moduleInfo.moduleSourceHash = newModuleHash;
        // Find the nearest boundary for the file that changed
        const moduleBoundary = this.getModuleBoundary(modulePath);
        // Notify the clients that a module has been updated
        this.emit('update', moduleBoundary);
    }
    /**
     * Identify which modulePath that represents the boundary of the changed module.
     * In the case of .html files, the module boundary is the template file itself. swapTemplate()
     *  is called and module swap is done.
     * In the case of .css files, the module boundary is the css file itself, swapStyle() is called
     *  and module swap is done.
     * In the case of .js files, the module boundary is the main entry file of the module, all the files
     *  reachable from the entry module are bundled and swapComponent invoked.
     *  Since JS hot swap is currently supporting components, this focused update of the main entry file
     *  is acceptable.
     * @param modulePath The module that changed
     * @return The modulePath of the boundary that needs to be updated.
     */
    getModuleBoundary(modulePath) {
        const { ext } = path.parse(modulePath);
        if (['.html', '.css'].includes(ext)) {
            return modulePath;
        }
        else if (['.js', '.ts'].includes(ext)) {
            const ownerModuleId = this.modulePathToModuleInfo[modulePath].ownerModuleId;
            const moduleMapEntry = this.moduleMap.get(ownerModuleId);
            if (moduleMapEntry) {
                const [namespace, name] = ownerModuleId.split('/');
                const { moduleParts } = moduleMapEntry;
                const entryFile = [
                    `${namespace}/${name}/${name}.js`,
                    `${namespace}/${name}/${name}.ts`,
                    `${namespace}/${name}/${name}.css`,
                ].find((entry) => Object.hasOwnProperty.call(moduleParts, entry));
                if (entryFile) {
                    return entryFile;
                }
                throw new Error(`[LWC Dev Server] Unable to determine entry file for path: ${modulePath}`);
            }
        }
        throw new Error(`[LWC Dev Server] Unrecognized file type or file not part of LWC module: ${modulePath}`);
    }
    /**
     * Fetch files and source related to the modulePath that is being hot loaded.
     * @param {string} name Name of the component being compiled
     * @param {string} namespace The namespace of the component
     * @param {string} modulePath The formatted path for the file being compiled.
     * @returns
     */
    getModuleFiles(name, namespace, modulePath) {
        const { ext } = path.parse(modulePath);
        const moduleDir = `${namespace}/${name}`;
        if (['.html', '.css'].includes(ext)) {
            const moduleFilePath = this.getFileByModulePath(modulePath);
            const fileEntry = path.relative(moduleDir, modulePath);
            // It can be assumed that the module file exists and the source can be read
            const src = fs.readFileSync(moduleFilePath, 'utf-8');
            return { [fileEntry]: src };
        }
        else if (['.js', '.ts'].includes(ext)) {
            const ownerModuleId = this.modulePathToModuleInfo[modulePath].ownerModuleId;
            const moduleMapEntry = this.moduleMap.get(ownerModuleId);
            return Object.keys(moduleMapEntry.moduleParts).reduce((acc, key) => {
                const part = path.relative(moduleDir, key);
                const partSrc = fs.readFileSync(moduleMapEntry.moduleParts[key].moduleFilePath, 'utf-8');
                acc[part] = partSrc;
                return acc;
            }, {});
        }
        throw new Error(`[LWC Dev Server] Unable to determine source for '${modulePath}'`);
    }
    close() {
        this.watcher.close();
    }
}
//# sourceMappingURL=index.js.map